# Generated by Django 2.2.10 on 2020-05-12 10:49

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("storage", "0006_assign_referenced_paths"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="referencedpath",
            name="file_storage",
        ),
    ]
